#ifndef _RNASTRUCT_H_
#define _RNASTRUCT_H_
#include "Util.h"
#include "MSA.h"


class segment;

/*************************enum ETYPE*********************************/
enum ETYPE{
	etype_unknown,
	etype_P0,
	etype_P1,
	etype_F,
	etype_C,
	etype_FM,
	etype_Fn, 
	etype_Fd, 
	etype_F2n,
	etype_F2d, 
	etype_M1, 
	etype_M2, 
	etype_M4
};
void showETYPE(const ETYPE & etype, ostream & fout = cout);
/*************************CLASS BP*********************************/
class BP{//base pair
	public: 
	int lp; //index of the left parenthesis
	int rp; //index of the right parenthesis
	BP() {lp=rp=0;}
	BP(const BP & in_bp) {
		lp = in_bp.lp; rp = in_bp.rp;
	}
	BP(int i1, int i2) {
		lp=i1; rp=i2;
	}
	void show(ostream & fout = cout) {
		fout<<"("<<lp<<" "<<rp<<")";
	}
	bool conflictwithBP(const BP & in_bp) {
		if(in_bp.lp <= lp && lp <= in_bp.rp && in_bp.rp <= rp) return true;
		if(lp <= in_bp.lp && in_bp.lp <= rp && rp <= in_bp.rp) return true;
		return false;
	}
};
bool operator == (const BP & in_bp1, const BP & in_bp2);
bool operator != (const BP & in_bp1, const BP & in_bp2);
bool operator < (const BP & in_bp1, const BP & in_bp2);
bool operator > (const BP & in_bp1, const BP & in_bp2);

class bplist{
	public:
	vector<BP> vbps;
	bplist(){}
	bplist(const vector<BP> & in_vbps) {
		vbps = in_vbps;
	}
	int size() {return sz(vbps);}
	void clear() {vbps.clear();}
	BP & operator [] (const int & i) {return vbps[i];}
	void insert(const BP & in_bp) {vbps.insert(vbps.begin(), in_bp);}
	void insert(const bplist & in_bpl) {vbps.insert(vbps.begin(), in_bpl.vbps.begin(), 
	     in_bpl.vbps.end());}
	void PB(const BP & in_bp) {vbps.PB(in_bp);}
	void PB(const bplist & in_bpl) {vbps.insert(vbps.end(), in_bpl.vbps.begin(),
	     in_bpl.vbps.end());}
//	BP pop() {BP bp = vbps[0]; vbps.erase(vbps.begin()); return bp;}
	BP pop_back() {BP bp = vbps[sz(vbps)-1]; vbps.erase(vbps.end()); return bp;}
	void show(ostream & fout = cout) {
		FOR(i,0,sz(vbps)){
			cout<<"bps ["<<i<<"/"<<sz(vbps)-1<<"]: "; vbps[i].show(); cout<<endl;
		}
	}
	string tostring(const int & len) {
		string str = ""; 
		str.insert(0,len,'.');
		FOR(i,0,sz(vbps)) {
			int a = vbps[i].lp; int b = vbps[i].rp;
			assert(str[a]=='.' && str[b]=='.');
			str[a] = '('; str[b] = ')';
		}
		return str;
	}	
	void output(ostream & fout, const int & len) {
		fout<<tostring(len)<<endl;
	}
};


class stem {
	public:
	BP outer; //for short (x=outer.lp, y=outer.rp, p=inner.lp, q=inner.rp)
	BP inner;
	stem(){outer.lp=outer.rp=inner.lp=inner.rp=-1;}
	stem(const int & x, const int & y, const int & p, const int & q) {
		outer.lp = x; outer.rp = y;
		inner.lp = p; inner.rp = q;
	}
	stem(BP bp1, BP bp2) {
		outer = bp1;
		inner = bp2;
	}
	stem(const stem & st) {
		outer = st.outer;
		inner = st.inner;
	}
	
	bool valid() {
		if(outer.lp<0 || outer.rp<0 || 
		   inner.lp<0 || inner.rp<0 || 
		   outer.lp>=inner.lp || inner.lp>=inner.rp || inner.rp>=outer.rp)
			return false;
		return true;
	}
	bool enclose(const stem & st) {//check if this stem totally enclose stem st \
i.e. ((((......(((((.....))))).........)))), the outer stem enclose the inner  \
stem
		return (inner.lp < st.outer.lp && inner.rp > st.outer.rp);
	}
	bool enclosedby(const stem & st) {
		return (st.inner.lp < outer.lp && outer.rp < st.inner.rp);
	}
	bool inward_extend(const stem & s);
	bool parallel_extend(const stem & s);
	bool prevparallel_extend(const stem & s);
	int inwhichseg(const int & query);
	int pairwithwhich(const int i) {
		if(i<outer.lp || i>outer.rp || (i>inner.lp && i<inner.rp)) 
			return -1;
		else 
			return (outer.lp+outer.rp-i);
	}
//	int len() {return (outer.rp-outer.lp+1);}
	int width() {return outer.rp-inner.rp+1;}
//	bool ispartof (const stem & st){//check if this stem is part of another stem
//		return ();
//	}
	//check if this stem is conflicting with in_bp?
	bool conflictwithBP(const BP & in_bp) {
		if (in_bp.rp < outer.lp || in_bp.lp > outer.rp) return false;
		if (in_bp.lp > inner.lp && in_bp.rp < inner.rp) return false;
		if (in_bp.lp < outer.lp && in_bp.rp > outer.rp) return false;
		return true;
	}
	bool conflictwith(stem st){
		if(st.outer.lp > outer.rp || st.outer.rp < outer.lp || 
		   (st.outer.lp > inner.lp && st.outer.rp < inner.rp) ||
		   (st.inner.lp < outer.lp && st.inner.rp > outer.rp))
		   return false;
		else return true;
	}
	bool containBP(const BP & in_bp) {
		if (in_bp.lp >= outer.lp && in_bp.rp <= outer.rp 
		 && in_bp.lp <= inner.lp && in_bp.rp >= inner.rp 
		 && in_bp.lp-outer.lp == outer.rp-in_bp.rp)
		 return true;
		return false;
	}
	bool containNT(const int & in_nt) {
		if (in_nt >= outer.lp && in_nt <= inner.lp) return true;
		if (in_nt >= inner.rp && in_nt <= outer.rp) return true;
		return false;
	}
	BP getBPcontainNT(const int & in_nt) {
		//return BP(in_nt, x) or BP(x, in_nt)
		if (in_nt>=outer.lp && in_nt<=inner.lp) return BP(in_nt, (outer.rp+outer.lp-in_nt));
		if (in_nt>=inner.rp && in_nt<=outer.rp) return BP((outer.rp+outer.lp-in_nt), in_nt);
	}
	void show(ostream & fout = cout) {
		//cout<<"["<<interval_s<<","<<interval_e<<"], 
		fout<<"("<<outer.lp<<","<<outer.rp<<") ~ ("<<inner.lp<<","<<inner.rp<<")";
	}
	int gethydrobonds(const string & seq) {
		int ret = 0; 
		int i = outer.lp; int j=outer.rp;
		while(i<=inner.lp) {
			if((seq[i] == 'A' && seq[j] == 'U') || (seq[i] == 'U' && seq[j] == 'A')) {
				ret += 2;
			}
			if((seq[i] == 'G' && seq[j] == 'C') || (seq[i] == 'C' && seq[j] == 'G')) {
				ret += 3;
			}
			if((seq[i] == 'G' && seq[j] == 'U') || (seq[i] == 'U' && seq[j] == 'G')) {
				ret += 1;
			}
			i++; j--;
		}
		return ret;
	}
	int gethydrobonds(MSA & in_msa) {
		int ret = 0; 
		FOR(n, 0, in_msa.N) {
			ret += gethydrobonds(in_msa[n]);
		}
		return ret;
	}
};

class stemlist {
	public:
	int L; 
	vector<stem> vstems;
	
	stemlist(){L = 0; vstems.clear();}
	stemlist(const string & seq);
	stemlist(vector< vector<bool> > & pi, MSA & in_msa);
	stemlist(const int & l, const double & cutoff, const vector< vector<double> > & b, vector< vector<int> > & pi);

	void clear() {vstems.clear(); }
	void show();
	int size() {return sz(vstems);}
	int getstem_index(const stem & in_st); //return index of in_st in vstems
	
	vector<stem> getstems_startfrom(int i);
	vector<stem> getstems_startfrom_within(int in_s, int in_e);
	vector<stem> getstems_endat(int in_e);
	vector<stem> getstems_startfrom_endat(int in_s, int in_e);
	vector<stem> getstems_within(int in_s, int in_e);

	vector<int> getstemindice_startfrom(int i);
	vector<int> getstemindice_startfrom_within(int in_s, int in_e);
	vector<int> getstemindice_endat(int in_e);
	vector<int> getstemindice_endat_within(int in_s, int in_e);
	vector<int> getstemindice_startfrom_endat(int in_s, int in_e);
	
	
	vector<int> getstemindice_enclosedby(int indx_s);
	vector<int> getstemindice_within(int in_s, int in_e);
	vector<int> getstemindice_inward_extend(int indx_s);
	vector<int> getstemindice_enclosedby_or_inward_extend(int indx_s);
	vector<int> getstemindice_prevparallel_extend(int indx_s);
	vector<int> getstemindice_prevsibling_or_prevparallel_extend(int indx_s); //return indices of stems that each stem t satisfies (ty==sx-1 or s and t overlap by a small fraction)
	vector<int> getstemindice_prevsibling_or_prevparallel_extend(const stem & in_st);
	
	stem getsubstem_startfrom_endat(const int & i, const int & j);
	vector<stem> getsubstems_enclosedby_or_inward_extend(const stem & s);
	vector<stem> getsubstems_between(stem & v, stem & w);
	vector<stem> getsubstems_within(const int & i, const int & j);
	vector<stem> getsubstems_startfrom_within(const int & i, const int & j);
	vector<stem> getsubstems_endat(const int & j);
	vector<stem> getsubstems_endat_within(const int & i, const int & j);

	vector<int> getFSet(const int & int_s, const int & in_e);

	bool conflictwith(stem st);
	bool containST(stem st);
	bool isRNAStructure();
	
	void insert(stem & in_st);
	void insert(stemlist & in_stl);
	void PB(stem & in_st);
	void PB(stemlist & in_stl);
	stem & operator [] (const int & i);
	stem pop() {stem st = vstems[0]; vstems.erase(vstems.begin()); return st;}
	void add(stem & in_st);
	stem findLS(const stem & in_st);
	stem findRS(const stem & in_st);
	stem findLS_4FM(const stem & in_st);
	stem findRS_4FM(const stem & in_st);
	stem findLS(const segment & in_sg); //return the right-most stem to the left of in_sg
	stem findRS(const segment & in_sg); //return the left-most stem to the right of in_sg
	stem inwhichST(const int & in_lp, const int & in_rp); //in which stack is the input base pair.
private:
	void construct_usingpi(vector< vector<bool> > & pi);
	void removestem_byhydrobonds(const string & seq);
	void removestem_byhydrobonds(MSA & in_msa);	
};

class segment{
	public: 
	int lp;  //index of the left parenthesis
	int rp;	//index of the right parenthesis
	segment(){}
	segment(int in_lp, int in_rp) {
		if(in_rp<in_lp) {cout<<"("<<in_lp<<","<<in_rp<<")"<<endl;}
		assert(in_rp>=in_lp);		
		lp=in_lp; rp=in_rp;
	}
	segment(BP bp) {assert(bp.lp<bp.rp); lp=bp.lp; rp=bp.rp;}
	bool overlap(stem st) {
		if(rp < st.outer.lp || lp > st.outer.rp ||
		   (st.inner.lp < lp && st.inner.rp > rp) )
		   return false;
		return true;
	}
	bool enclose(const stem & st) {
		if(lp <= st.outer.lp && rp >= st.outer.rp) 
			return true;
		return false;
	}
	bool valid() {
		if(lp>=0 && rp>=0 && lp<=rp) return true;
		else return false;
	}
	bool operator == (const segment & sg) {
		return (lp ==  sg.lp && rp == sg.rp);
	}
	bool operator != (const segment & sg) {
		return (lp != sg.lp || rp != sg.rp);
	}
	bool operator < (const segment & sg) {
		return (lp < sg.lp || (lp==sg.rp && rp>sg.rp));
	}
	bool operator > (const segment & sg) {
		return (lp > sg.lp || (lp==sg.lp && rp<sg.rp));
	}
	void show(ostream & fout = cout) {
		fout<<"["<<lp<<","<<rp<<"]";
	}
};

class segmentT:public segment{//segment with type
	public: 
	ETYPE etype;
	segmentT(segment in_sg, ETYPE in_type = etype_unknown) {
		lp = in_sg.lp; rp = in_sg.rp; etype = in_type;
	}
	segmentT(int in_lp, int in_rp, ETYPE in_type = etype_unknown) {
		lp = in_lp; rp = in_rp; etype = in_type;
	}
	segmentT(BP in_bp, ETYPE in_type = etype_unknown) {
		lp = in_bp.lp; rp = in_bp.rp; etype = in_type;			
	}
	void show(ostream & fout = cout) {
		fout <<"["<<lp<<", "<<rp<<"]_";
		showETYPE(etype, fout);
	}
};

class seglist{
	public:
	vector<segment> vsegs;
	seglist(){}
	seglist(const vector<segment> & in_vsg) {
		vsegs = in_vsg;
	}
	int size() {return sz(vsegs);}
	void clear() {vsegs.clear();}
	bool overlap(stem st) {
		FOR(i,0,sz(vsegs)) {
			if(vsegs[i].overlap(st)) return true;
		}
		return false;
	}
	segment & operator [] (const int & i) {return vsegs[i];}
	void insert(const segment & in_sg) {vsegs.insert(vsegs.begin(), in_sg);}
	void insert(const seglist & in_sgl) {vsegs.insert(vsegs.begin(), 
	     in_sgl.vsegs.begin(), in_sgl.vsegs.end());}
	void PB(const segment & in_sg) {vsegs.PB(in_sg);}
	void PB(const seglist & in_sgl) {vsegs.insert(vsegs.end(), 
	     in_sgl.vsegs.begin(), in_sgl.vsegs.end());}
	segment pop_back() {segment sg = vsegs[sz(vsegs)-1]; vsegs.erase(vsegs.end()); return sg;}
	segment findRS(const segment & in_sg);
	void show(ostream & fout = cout) {
		FOR(i,0,sz(vsegs)){
			fout<<"segs ["<<i<<"/"<<sz(vsegs)-1<<"] :";
			vsegs[i].show();
			fout<<endl;
		}
	}
};

class seglistT{
	public:
	vector<segmentT> vsegs; 
	seglistT(){}
	seglistT(const seglistT & in_sgl) {
		vsegs = in_sgl.vsegs;
	}
	seglistT(const vector<segmentT> & in_vsg) {
		vsegs = in_vsg;
	}
	int size() {return sz(vsegs);}
	void clear() {vsegs.clear();}
	bool overlap(stem st) {
		FOR(i,0,sz(vsegs)) {
			if(vsegs[i].overlap(st)) return true;
		}
		return false;
	}
	segmentT & operator [] (const int & i) {return vsegs[i];}
	void insert(const segmentT & in_sg) {vsegs.insert(vsegs.begin(), in_sg);}
	void insert(const seglistT & in_sgl) {vsegs.insert(vsegs.begin(), 
	     in_sgl.vsegs.begin(), in_sgl.vsegs.end());}
	void PB(const segmentT & in_sg) {vsegs.PB(in_sg);}
	void PB(const seglistT & in_sgl) {vsegs.insert(vsegs.end(), 
	     in_sgl.vsegs.begin(), in_sgl.vsegs.end());}
	segmentT pop_back() {segmentT sg = vsegs[sz(vsegs)-1]; vsegs.erase(vsegs.end()); return sg;}
	segmentT findRS(const segmentT & in_sg);
	void show() {
		FOR(i,0,sz(vsegs)){
			cout<<"segs ["<<i<<"/"<<sz(vsegs)-1<<"]: ";
			vsegs[i].show();
			cout<<endl;
		}
	}
};


class RStack{
	public:
	vector<stem> vstems; //all stacks in the structure
	int L; //the length of the sequence 
	RStack(){L=0;vstems.clear();}
	RStack(const vector<stem> in_vst, const int & in_len) {vstems = in_vst;L = in_len;}
	RStack(const string & str);

	bool ismatch(vector<  vector<bool> > & ali_pi);
	string tostring();
	int size() {return sz(vstems);}
	void show(){cout<<tostring()<<endl;}
	void show2(){
		cout<<"The stems in RStack is "<<endl;
		FOR(i,0,sz(vstems)) {
			vstems[i].show();
			cout<<endl;
		}
	}

	bool containST(const stem & st);//whether this instance contains stack st?
	int  inwhichST(const int & query); //in which stack is query.
	bool isLocOpt(const stemlist & in_stl);
	vector<RStack> getneighbor_delstack();
	vector<int> findConflictingST(stem & sti);//return indices of stems that conflict with sti
};

class PartStruct{
	public: 
	seglist sgl;
	bplist bpl;
	
	PartStruct(){}
	PartStruct(const PartStruct & phi) {
		clear();
		sgl.insert(phi.sgl);
		bpl.insert(phi.bpl);
	}
	PartStruct(const seglist & in_sgl, const bplist & in_bpl){
		sgl = in_sgl; bpl = in_bpl;
	}
	void clear() {sgl.clear(); bpl.clear();}
	void show() {
		cout<<"PartStruct segments:";
		FOR(i,0,sz(sgl)) {
			sgl[i].show(); cout<<", ";
		}		
		cout<<endl<<"PartStruct base pairs:";
		FOR(i,0,sz(bpl)){
			bpl[i].show(); cout<<", ";
		}
		cout<<endl;
	}
};

class PartStructT{
	public:
	seglistT sgl;
	bplist bpl;
	
	PartStructT(){}
	PartStructT(const PartStructT & phi) {
		clear();
		sgl.insert(phi.sgl);
		bpl.insert(phi.bpl);
	}
	
	PartStructT(const seglistT & in_sgl, const bplist & in_bpl){
		sgl = in_sgl; bpl = in_bpl;
	}
	
	void clear() {sgl.clear(); bpl.clear();}
	void show() {
		cout<<"PartStructT segments:";
		FOR(i,0,sz(sgl)) {
			sgl[i].show(); cout<<", ";
		}		
		cout<<endl<<"PartStructT base pairs:";
		FOR(i,0,sz(bpl)){
			bpl[i].show(); cout<<", ";
		}
		cout<<endl;
	}
};

class PartStackStruct{
	public: 
	seglistT sgl;
	stemlist stl;
	int elphi;
	string log;

	PartStackStruct(){elphi=0; log="";}
	PartStackStruct(const PartStackStruct & phi) {
		clear();
		sgl.vsegs.insert(sgl.vsegs.begin(), phi.sgl.vsegs.begin(), phi.sgl.vsegs.end());
		stl.vstems.insert(stl.vstems.begin(), phi.stl.vstems.begin(), phi.stl.vstems.end());
		elphi=phi.elphi; 
		log=phi.log;
	}
	PartStackStruct(const seglistT & in_sgl, const stemlist & in_stl, 
			const int in_elphi = 0, const string & in_log=""){
		sgl.vsegs.insert(sgl.vsegs.begin(), in_sgl.vsegs.begin(), in_sgl.vsegs.end());
		stl.vstems.insert(stl.vstems.begin(), in_stl.vstems.begin(), in_stl.vstems.end());
		elphi = in_elphi; 
		log = in_log;
	}
	PartStackStruct(const vector<segmentT> & in_vsg, const vector<stem> & in_vst) {
		sgl.vsegs = in_vsg; stl.vstems= in_vst;
	}
	void clear() {sgl.clear(); stl.clear(); elphi=0; log="";}
	string tostring(const int & length) {
		assert(sz(sgl)==0);
		string str = ""; str.insert(0, length, '.');
		FOR(i,0,sz(stl)) {
			stem st = stl[i];
			int tx = st.outer.lp; int ty = st.outer.rp;
			int tp = st.inner.lp; int tq = st.inner.rp;
			FOR(j, tx, tp+1) {
				str[j] = '('; str[tx+ty-j] = ')';
			}
		}
		return str; 
	}
	void output(const int & length, ostream & fout = cout) {
		fout<<tostring(length)<<endl;
	}
	void show(ostream & fout = cout) {
		fout<<"PartStackStruct segmentTs:";
		FOR(i,0,sz(sgl))
			sgl.vsegs[i].show(fout);
//			fout<<"["<<sgl.vsegs[i].lp<<", "<<sgl.vsegs[i].rp<<"]_";showETYPE(sgl.vsegs
		fout<<endl<<"PartStackStruct stacks:";
		FOR(i,0,sz(stl)){
			stl.vstems[i].show(fout);
		}
		fout<<endl;
	}
	bool isLocOpt(stemlist & in_stl);
};


class RNA{
	public:
	string str;
	double e; //free energy
	double c; //covariance value
	RNA(){};
	~RNA(){};
	RNA(const string & in_str, const double & in_e, const double & in_c) {
		str = in_str; e = in_e; c = in_c;
	}
};


/* functions for classes */
bool operator == (const stem & st1, const stem & st2);
bool operator != (const stem & st1, const stem & st2);
bool operator <  (const stem & st1, const stem & st2);
bool operator >  (const stem & st1, const stem & st2);

/* functions used in common */
int getindx(const vector<int>& vi, const int& query, const int & startpos=0, const int & direction=1);
int getindx(const vector<BP> & in_vbp, const BP & in_bp, const int & startpos=0, const int & direction=1);//return the index of in_bp in in_vbp
//int getindx(const vector<Action> & in_va, const Action & in_a, const int & startpos=0, const int & direction=1);//return the index of the first in_a in in_va
int getindx(const vector<stem> & in_vst, const stem & in_st, const int & startpos=0, const int & direction=1);
vector<BP> s2v(const string & str);//given a string representation of a strucuture, return vbp
string v2s(const int & l, const vector<BP> & in_vbp);//given indice of the left/right parenthesis and the length of the RStruct, return s;
vector<int> getunp(const int & l, const vector<BP> & in_vbp);//return indices of unpaired bases
vector<int> getunp(const string & str);//return indices of unpaired bases
bool descendingsort(double d1, double d2);
bool descendingsortBP(BP b1, BP b2);
bool descendingsortST(stem st1, stem st2);
bool ascendingsort(double d1, double d2);
bool ascendingsortBP(BP b1, BP b2);
bool ascendingsortST(stem st1, stem st2);
bool mycheckclash(const BP & in_bp, const vector<BP>& in_vbp);

bool descendingsortRNAbyE(RNA rna1, RNA rna2);
bool ascendingsortRNAbyE(RNA rna1, RNA rna2);
bool descendingsortRNAbyC(RNA rna1, RNA rna2);
bool ascendingsortRNAbyC(RNA rna1, RNA rna2);


#endif
