#include "Util.h"

void readin(ifstream & fin, string & seq) {
	char buf[MAXCHARNUM];
	fin.getline(buf, MAXCHARNUM);
	seq = string (buf);
}

//check if a file exists or not
bool fexists(const char *filename) {
  ifstream ifile(filename);
  return ifile;
}

//Extended on May 5 for multiple sequence alignment (be able to process gap)
string dna2rna(const string & seq) {
	string ret = "";
	FOR(i,0,sz(seq)){
		if(seq[i] == 'T'|| seq[i] == 't' || seq[i] == 'U' || seq[i] == 'u') {
			ret += "U";
		} else if (seq[i] == 'A' || seq[i] == 'a') {
			ret += "A";
		} else if (seq[i] == 'G' || seq[i] == 'g') {
			ret += "G";
		} else if (seq[i] == 'C' || seq[i] == 'c') {
			ret += "C";
		} else if (seq[i] == '-') {
			ret += "-";
		} else {
			cerr<<"Unknown character "<<seq[i]<<" found in an RNA sequence: "<< 
			  seq<< endl;
			ret += "-";
//			exit(0);
		}
	}
	return ret;
}

double max(const double & d1, const double & d2){
	return (d1>d2)?(d1):(d2);
}

double min(const double & d1, const double & d2){
	return (d1>d2)?(d2):(d1);
}

int max(const int & d1, const int & d2){
	return (d1>d2)?(d1):(d2);
}

int min(const int & d1, const int & d2){
	return (d1>d2)?(d2):(d1);
}

string lowercase(const string & str) {
	string ret = ""; 
	FOR(i,0,sz(str)) {
		char c = tolower(str[i]);
		ret.insert(ret.end(), 1, c);
	}
	return ret;
}

string uppercase(const string & str) {
	string ret = ""; 
	FOR(i,0,sz(str)) {
		char c = toupper(str[i]);
		ret.insert(ret.end(), 1, c);
	}
	return ret;
}

//trim the following characters from both ends 
string trim(const string & str) {
	string ret = "";
	int start = 0; int end = sz(str)-1;
	FOR(i,0,sz(str)) {
		start = i;
		if(str[i] == '\t' || str[i] == '\n' || str[i] == ' ') {
		} else {	
			break;
		}
	}
	for(int i = sz(str)-1; i>=0; i--) {
		end = i-1;
		if(str[i] == '\t' || str[i] == '\n' || str[i] == ' ') {
		} else {
//			if(str[i]!='A' && str[i]!='T' && str[i]!='G' && str[i]!='C' && str[i]!='U' && str[i]!='-')
//				cout<<"seq=["<<str<<"]"<<int(str[i])<<endl;
			break;
		}
	}
	if (start <= end) 
		ret = str.substr(start, (end-start+1));
//	cout<<"start="<<start<<", end="<<end<<", ret=["<<ret<<"]"<<endl;
	return ret;
}

void parseclustalline(const string & str, string & first, string & second) {
	first = second = "";
	bool readfirst = true;
	FOR(i,0,sz(str)) {
		if(str[i]!=' ' && str[i]!='\t') {
			if(readfirst) {
				first.insert(first.end(), str[i]);
			} else {
				second.insert(second.end(), str[i]);
			}
		} else {//once a space or a \t is encountered, read the second
			readfirst = false;
		}
	}
}

//if a==b, return 1, else return 0
int cmp(const char & a, const char & b) {
	if(a==b) return 1; 
	else return 0;
}

//Extended on May 5 for multiple sequence alignment (be able to process gap)
bool match(const char & x , const char & y, bool allowgu) {
	if(x=='T' && y=='A') return true;
	if(x=='U' && y=='A') return true;
	if(x=='G' && y=='C') return true;
	if(x=='C' && y=='G') return true;
	if(x=='A' && y=='T') return true;
	if(x=='A' && y=='U') return true;
//modified by Yuan Li April 10 2012
//	if(x=='-' && y=='-') return true;
	if(allowgu==true) {
		if(x=='G' && y=='U') return true;
		if(x=='U' && y=='G') return true;
	}
	return false;
}

int hydrogenbond(const char & x, const char & y) {
	if(x=='T' && y=='A') return 2;
	if(x=='U' && y=='A') return 2;
	if(x=='G' && y=='C') return 3;
	if(x=='C' && y=='G') return 3;
	if(x=='A' && y=='T') return 2;
	if(x=='A' && y=='U') return 2;
	if(x=='-' && y=='-') return 0;
	if(x=='G' && y=='U') return 1;
	if(x=='U' && y=='G') return 1;
	return 0;
}


char getcomplementchar(const char & x) {
	switch(x) {
	case 'A': return 'U';
	case 'U':
	case 'T': return 'A';
	case 'G': return 'C';
	case 'C': return 'G';
	default : {cerr<<"The input nucleotide is "<<x<<endl; exit(0);}
	}
	return 'A';
}

bool eq(const double & x, const double & y) {
	return (x-y<=PRECISION && x-y>=-PRECISION);
}

//random number generator
int myrand(int n) {
	if(n==0) return 0;
   return rand() % n;
}

void show(vector<vector<bool> > & vvb, ostream & fout) {
	FOR(i, 0, sz(vvb)) {
		FOR(j, 0, sz(vvb[i])) {
			if(vvb[i][j]) fout <<".";
			else fout<<" ";
		}
		if(i!=sz(vvb)-1) fout<<endl;
	}
	fout<<endl;	
}

void showUpTriMatrix(vector<int> & vi, const int & blocksz, ostream & fout) {
	int length = blocksz;
	int count = 0;
	fout.precision(5);
	FOR(i, 0, blocksz) {
		FOR(j, 0, blocksz) {
			int ij = (2*blocksz-i+1)*i/2+j-i;
			if(i>j) fout<<"NAN ";
			else if(vi[ij]==OO) 
				fout<<"INF ";
			else fout<<vi[ij]<<" ";
		}
		fout<<endl;
	}
	fout<<endl;
}

void show(vector<int> & vi, ostream & fout) {
	FOR(i, 0, sz(vi)) {
		fout.precision(4);
		if(vi[i]==OO) 
			fout<<"INF ";
		else fout<<vi[i]<<" ";
		if(i!=sz(vi)-1) fout<<"\t";
	}
	fout<<endl;
}

void show(vector<vector<int> > & vvi, ostream & fout) {
	FOR(i, 0, sz(vvi)) {
		FOR(j, 0, sz(vvi[i])) {
			fout<<vvi[i][j]<<" ";			
		}
		fout<<endl;
	}
	fout<<endl;
}

void show(vector<double> & vd, ostream & fout) {
	FOR(i, 0, sz(vd)) {
		fout.precision(4);
		if(eq(vd[i], OO)) 
			fout<<"INF";
		else if(eq(vd[i], 0))
			fout<<"0  ";
		else fout<<vd[i];

		if(i!=sz(vd)-1) fout<<"\t";
	}
	fout<<endl;
}

void show(vector<vector<double> > & vvd, ostream & fout) {
	if(sz(vvd)==0) {
		fout<<endl;
		return ;
	}
	FOR(i,0,sz(vvd)) {
		FOR(j,0,sz(vvd[i])) {
			fout.precision(4);
			if(eq(vvd[i][j], OO)) 
				fout<<"INF";
			else if(eq(vvd[i][j], 0))
				fout<<"0  ";
			else fout<<vvd[i][j];
			if(j!=sz(vvd[i])-1) fout<<"\t";
		}
		fout<<endl;
	}
}

//The length of in_str should be equivalent to the number of 1s in gap2ngap,
//in_str should be obtained through removing all the columns that are 0 in the gap2ngap vector from the originial string. 
string reverse_gap2ngap(const string & in_str, const vector<int> & gap2ngap) {
	string ret;
	int index = 0;
	FOR(j, 0, sz(gap2ngap)) {
		if(gap2ngap[j]==0) {
			ret += "-";
		} else {
			ret += in_str[index];
			index++;
		}
	}
	assert(index==sz(in_str));
	assert(sz(ret)==sz(gap2ngap));
	return ret;
}














